# -*- coding: utf-8 -*-
"""
Created on Mon Nov  7 16:55:11 2022

@author: Frank
"""

import imodmodel
import emfile
import numpy as np
from scipy.spatial.transform import Rotation as R
import scipy.io
from pathlib import Path
from glob import glob
from zipfile import ZipFile

from .classes import SubTomogram, Particle

# readin
TEMP_FILE_DIR = "static/"


def em_format(particle):
    rx, ry, rz = particle.direction
    # convert orientation vector into euler angles
    # PlaceObject uses "zxz" euler angles, but saved in the order "zzx"
    rotation_matrix = np.array([[0, 0, rx], [0, 0, ry], [0, 0, rz]])
    euler = R.from_matrix(rotation_matrix).as_euler("xzx", degrees=True)
    euler_formatted = [euler[0], euler[2], euler[1]]
    # print("")
    return [
        particle.cc_score,
        0.0,
        0,
        0,
        0,
        0,
        0,
        *particle.position,
        0,
        0,
        0,
        0,
        0,
        0,
        *euler_formatted,
        1,
    ]


def modify_emc_mat(
    subtomo_dict: dict, out_path: str, inp_path: str, keep_selected: bool
):
    output_data = dict()
    for tomo_id in subtomo_dict.keys():
        print(tomo_id)
        table_rows = list()
        try:
            mat_out = scipy.io.loadmat(inp_path, simplify_cells=True, mat_dtype=True)
            mat_inp = mat_out["subTomoMeta"]["cycle000"]["geometry"]
        except:
            print("Unable to open original matlab file, was it moved?")
            return ""
        mat_table = mat_inp[tomo_id]
        if keep_selected:
            for particle_id in subtomo_dict[tomo_id].selected_particle_ids():
                table_rows.append(mat_table[particle_id])
        else:
            for particle_id in subtomo_dict[tomo_id].unselected_particle_ids():
                table_rows.append(mat_table[particle_id])
        output_data[tomo_id] = table_rows

    try:
        mat_out["subTomoMeta"]["cycle000"]["geometry"] = output_data
        scipy.io.savemat(out_path, mdict=mat_out)
    except:
        print("Unable to save file")

def write_emfile(subtomo_dict: dict, out_suffix: str, keep_selected: bool):
    for skey, tomo in subtomo_dict.items():
        filename = "{0}_{1}{2}".format(skey, out_suffix, ".em")
        good_particles = tomo.auto_cleaned_particles
        em_list = np.array([[em_format(particle) for particle in good_particles]])
        emfile.write(TEMP_FILE_DIR + filename, em_list, overwrite=True)


def zip_files(final_filename: str, extension_to_zip: str):
    archive_name = "{}.zip".format(final_filename)
    filenames = glob("*.{}".format(extension_to_zip))

    with ZipFile(archive_name, mode="w") as zp:
        for filename in filenames:
            zp.write(filename)

    return archive_name


def read_imod(filename: str):
    """


    Parameters
    ----------
    filename : str
        imod filename

    Returns
    -------
    3xN np array (x, y, z)
        e.g.
        [[27.4   99.4   3.6]
         [301.2  38.2   8.1]
         ...
         [43.1   99.0  12.3]
         [88.2   21.1  33.8]]

    """
    imod_inp = imodmodel.read(filename)
    return np.transpose(np.array([imod_inp[q] for q in ["x", "y", "z"]]))


def read_emfile(filename: str):
    header, em_inp = emfile.read(filename)
    print(em_inp)


def append_filename(filename, suffix="out"):
    p = Path(filename)
    return "{0}_{1}{2}".format(p.stem, suffix, p.suffix)


def mat_to_subtomos(filename, cycle="cycle000", num_images=-1):
    """
    Read subtomograms from a .mat file generated by emClarity

    Parameters
    ----------
    filename : String
        Path to .mat file
    cycle : string, optional
        Key of cycle to read from database
        Defaults to cycle000.
    num_images : int
        Number of tomograms to read
        Defaults to all (-1).

    Returns
    -------
    subtomograms : dict {string : SubTomogram}
        Dictionary of read subtomgrams
        If file is unreadable, returns 0

    """
    geom_key = "geometry" if cycle=="cycle000" else "Avg_geometry"
    
    try:
        full_geom = scipy.io.loadmat(filename, simplify_cells=True)[
            "subTomoMeta"
        ][cycle][geom_key]
    except:
        print("Mat file {} unreadable".format(filename))
        return
    
    subtomograms = dict()
    for idx, (gkey, geom) in enumerate(full_geom.items()):
        if idx == num_images:
            break
        print(gkey)
        subtomo = SubTomogram(gkey)
        pdata = [[row[0], row[10:13], row[22:25]] for row in geom]
        particles = Particle.from_array(pdata, subtomo)
        subtomo.set_particles(particles)
        #print(subtomo.name, len(subtomo.all_particles))
        subtomograms[gkey] = subtomo
    return subtomograms
# data=0


# print("header", header)
